# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
  
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
  
#       http://www.apache.org/licenses/LICENSE-2.0
  
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import boto3
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt


class Metrics():
    """Return metrics for a given Amazon Lookout for Vision project.

        Attributes:
        project_name    The name of the Amazon Lookout for Vision project.
        lv              The Amazon Lookout for Vision boto3 client.

    """

    def __init__(self, project_name):
        """Creates a Metrics object f°or Amazon Lookout for Vision.
        It can retrieve key metrics, namely precision, recall and f1 score
        from a given Amazon Lookout for Vision project.

        Technical documentation on how Amazon Lookout for Vision works can be
        found at: https://aws.amazon.com/lookout-for-vision/

        Args:
            project_name (str): Name of the Amazon Lookout for Vision to interact with.

        """
        super(Metrics, self).__init__()
        self.project_name = project_name
        self.lv = boto3.client("lookoutvision")

    def describe_model(self, model_version="1"):
        """Gets and outputs the metrics for a given model version of your project.

        Args:
            model_version (str): The version of the model that should be describedmodel_version

        Returns:
            pandas.core.frame.DataFrame: a DataFrame with the metadata of your model

        """
        status = self.lv.list_models(
            ProjectName=self.project_name
        )["Models"]
        model = [x for x in status if x["ModelVersion"] == model_version]
        if model[0]["Status"] not in ["TRAINED", "HOSTED"]:
            print("Warning: Your model needs to be in TRAINED or HOSTED state!")
            return pd.DataFrame([None])
        else:
            return pd.DataFrame(model[0])

    def describe_models(self):
        """Gets and outputs all metrics of your project.

        Args:
            None

        Returns:
            pandas.core.frame.DataFrame: a DataFrame with the metadata of all models

        """
        models = self.lv.list_models(
            ProjectName=self.project_name
        )["Models"]
        output = pd.DataFrame()
        for model in models:
            if model["Status"] not in ["TRAINED", "HOSTED"]:
                print("Warning: Your model needs to be in TRAINED or HOSTED state! Skipping model version {}".format(model["ModelVersion"]))
            else:
                output = pd.concat(objs=[output, pd.DataFrame(model)], axis=0)
        return output

    def compare_models(self, figsize=(8, 6)):
        """Visualizes the metrics of all models against each other

        Args:
            figisze (tupel): a tupel for (width, length)

        Returns:
            None

        """
        models = self.describe_models()
        plt.figure(figsize=figsize)
        ax = sns.barplot(x="ModelVersion", y="Performance",
                         hue=models.index, data=models)
        plt.show()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
  
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
  
#       http://www.apache.org/licenses/LICENSE-2.0
  
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os
import numpy as np
import boto3
from skimage import io, img_as_ubyte
from skimage.transform import rescale, resize
import math
import random
from multiprocessing.pool import ThreadPool


class Image():
    """Helper class to assist with image handling.

    This class can help you check your local image sizes, resize images
    check if they are compliant and help you upload local images
    into your S3 bucket.

    Attributes:
    s3              The Amazon S3 boto3 client.

    """

    def __init__(self):
        """Creates a Image helper object for Amazon Lookout for Vision.
        It will assist you in checking image sizes, complying with the
        Amazon Lookout for Vision image boundaries, rescaling and uploading
        local images to S3.

        Technical documentation on how Amazon Lookout for Vision works can be
        found at: https://aws.amazon.com/lookout-for-vision/

        Args:
            None

        """
        super(Image, self).__init__()
        self.s3 = boto3.client("s3")

    @classmethod
    def __image_size_checker(self, image_type="good", verbose=False):
        """Internal image size checker method.
        It will count the number of images that are compliant with
        Amazon Lookout for Vision.

        Args:
            image_type (str): can either be "good" or "bad".
                Make sure you have a local folders named "good"
                and "bad" with the respective images.
            verbose (bool): Decide to include image metadata in output.

        Returns:
            json: a JSON object with the meta information on the images.

        """
        # List all files and count the number of images.
        # Also set:
        # - internal counter 'cnt'
        # - minimal image size 'img_min'
        # - maximal image size 'img_max'
        # - image_info as the JSON object to collect image metadata
        files = os.listdir(image_type)
        no_of_images = len(files)
        cnt, img_min, img_max, image_info = 0, 64, 4096, {}
        # For each file...
        for file in files:
            # ...set the path and read the image in
            path = "{}/{}".format(image_type, file)
            if not (".jpeg" in path or ".jpg" in path or ".png" in path):
                print("The following image is not compliant: {}".format(path))
                continue
            image = io.imread(path)
            # Check the image min and max size and
            # log the shape
            min_size = min(image.shape[:2])
            max_size = max(image.shape[:2])
            image_info[path] = image.shape
            # If image complies cnt += 1
            if min_size >= img_min and max_size <= img_max:
                cnt += 1
            else:
                # if not log the smalles/largest image found
                img_min = min(img_min, min_size)
                img_max = max(img_max, max_size)
        # Check if all images are compliant
        compliant = True if cnt == no_of_images else False
        # Set output
        response = {
            "no_of_images": no_of_images,
            "compliant_images": cnt,
            "compliant": compliant
        }
        # Add some metadata if image is not compliant
        if not compliant:
            response["min_size"] = img_min
            response["max_size"] = img_max
        # Add more metadata
        if verbose:
            response["image_metadata"] = image_info
        return response

    @classmethod
    def __image_shape_checker(self, image_type="good", verbose=False):
        """Internal image shape checker method.
        This method can validate if all images have the same shape
        or if they need adjustment before Amazon Lookout for Vision
        can train a model.

        Args:
            image_type (str): can either be "good" or "bad".
                Make sure you have a local folders named "good"
                and "bad" with the respective images.
            verbose (bool): Decide to include image metadata in output.

        Returns:
            json: a JSON object with the meta information on the images.

        """
        # List all files and count the number of images.
        # Also set:
        # - internal counter 'cnt'
        # - placeholder for image shape 'shape'
        # - image_info as the JSON object to collect image metadata
        files = os.listdir(image_type)
        no_of_images = len(files)
        cnt, shape, image_info = 1, (0, 0, 0), {}
        value = 1e6
        # For each image...
        for file in files:
            # ...set path and read image
            path = "{}/{}".format(image_type, file)
            if not (".jpeg" in path or ".jpg" in path or ".png" in path):
                print("The following image is not compliant: {}".format(path))
                continue
            image = io.imread(path)
            # For the first image set 'shape' to the images's shape
            if shape == (0, 0, 0):
                shape = image.shape
            else:
                # Check how often images have the same shape
                if shape == image.shape:
                    cnt += 1
                else:
                    # If images found that are smaller than the other
                    # set the new shape value here (that way you can
                    # make sure to know what is the scaling factor to
                    # downscale images to):
                    minvalue = min(shape[:-1])
                    maxvalue = max(shape[:-1])
                    if shape[0] > shape[1] and maxvalue/minvalue < value:
                        value = maxvalue/minvalue
                        shape = image.shape
            # Log shape in output
            image_info[path] = image.shape
        # Set status if all images have the same size
        status = "Image sizes are equal!" if cnt == no_of_images else "Downsize images!"
        # Populate response
        response = {
            "no_of_images": no_of_images,
            "compliant": cnt,
            "status": status,
            "min_image_shape": shape
        }
        # Log more if asked for:
        if verbose:
            response["image_metadata"] = image_info
        return response

    @classmethod
    def __check_image_number(self, no_of_files):
        """This method will check the test images, if the total number overceed 5000.
        This is roughly based on the S3 API throttle limit:
        https://docs.aws.amazon.com/AmazonS3/latest/userguide/optimizing-performance.html

        Args:
            no_of_files (int): count of number of files

        Returns:
            json: a JSON object with the meta information on the images.

        """
        if no_of_files > 5000:
            print("Warning: You have too many images, this will slow down the upload process!")
            print("Images will be uploaded. If you want to speed up the process please use <5000 images.")
        return True

    def __upload_s3(self, filename, bucket, fname):
        """This method is used to do the upload to s3

        Args:
            filename (str): filname to upload
            bucket (str): S3 bucket
            fname (str): S3 object key

        Returns:
            None
        """
        self.s3.upload_file(
            Filename=filename,
            Bucket=bucket,
            Key=fname)

    @classmethod
    def __is_compliant(self, response):
        """Helper to return if all images are compliant.

        Args:
            response (json): Returned object from __image_size_checker

        Returns:
            bool: True or False (compliant or not compliant)

        """
        good = response["good"]["compliant"]
        bad = response["bad"]["compliant"]
        return good and bad

    @classmethod
    def __is_too_small(self, response):
        """Helper to return if any image used is too small.

        Args:
            response (json): Returned object from __image_size_checker

        Returns:
            bool: True or False

        """
        compliant = self.__is_compliant(response=response)
        if not compliant:
            return min(response["good"]["min_size"], response["bad"]["min_size"]) < 64
        return False

    @classmethod
    def __is_too_large(self, response):
        """Helper to return if any image used is too large.

        Args:
            response (json): Returned object from __image_size_checker

        Returns:
            bool: True or False

        """
        compliant = self.__is_compliant(response=response)
        if not compliant:
            return max(response["good"]["max_size"], response["bad"]["max_size"]) > 64
        return False

    def check_image_sizes(self, prefix="", verbose=False):
        """Check your image sizes with this function.

        Args:
            verbose (bool): Decide to include image metadata in output.

        Returns:
            json: Metadata on the image and their compliance.

        """
        # If you don't have folders "good" and "bad" this libary
        # will not work:
        folders = os.listdir()
        if "{}good".format(prefix) not in folders and "{}bad".format(prefix) not in folders:
            print(
                "Error: Methods requires folders good/ and bad/ with images in this location!")
            return {}
        # For each folder set a response
        response = {}
        for path in ["{}good".format(prefix), "{}bad".format(prefix)]:
            response[path] = self.__image_size_checker(
                image_type=path, verbose=verbose)
        return response

    def check_image_shapes(self, prefix="", verbose=False):
        """Check your image sahpes with this function.

        Args:
            verbose (bool): Decide to include image metadata in output.

        Returns:
            json: Metadata on the image and their compliance.

        """
        # If you don't have folders "good" and "bad" this libary
        # will not work:
        folders = os.listdir()
        if "{}good".format(prefix) not in folders and "{}bad".format(prefix) not in folders:
            print(
                "Error: Methods requires folders {}good/ and {}bad/ with images in this location!".format(prefix, prefix))
            return {}
        # For each folder set a response
        response = {}
        for path in ["{}good".format(prefix), "{}bad".format(prefix)]:
            response[path] = self.__image_shape_checker(
                image_type=path, verbose=verbose)
        good = response["{}good".format(prefix)]["min_image_shape"]
        bad = response["{}bad".format(prefix)]["min_image_shape"]
        if max(good[:-1])/min(good[:-1]) < max(bad[:-1])/min(bad[:-1]):
            response["shape_recommendation"] = good
        else:
            response["shape_recommendation"] = bad
        return response

    def rescale(self, prefix="rescaled_", force_rescale=False):
        """This method will rescale all your images to a common size.
        This function will automatically determine if your images comply
        and if they need rescaling. It will only rescale if your images
        are not of same size or violate the boundaries of Amazon Lookout for Vision.

        Args:
            prefix (str): The prefix for the new folder with your resized images.
                Note: if set to "" this function will overwrite your original images.
            force_rescale (bool): force to rescale even for compliant images.

        Returns:
            json: Object with the new local image paths.

        """
        # Factor = 1 means images need no rescaling
        # Also, if the function returns the default output all images are
        # good as they are:
        factor = 1
        output = {
            "{}good".format(prefix): "Ok",
            "{}bad".format(prefix): "Ok"
        }
        # If you don't have folders "good" and "bad" this libary
        # will not work:
        folders = os.listdir()
        if "good" not in folders and "bad" not in folders:
            print(
                "Error: Methods requires folders good/ and bad/ with images in this location!")
            return {}
        shapes = self.check_image_shapes()
        shape_rec = shapes["shape_recommendation"]
        if max(shape_rec) > 4096 or min(shape_rec[:-1]) < 64:
            print("Warning: Your images shapes imply that rescaling will lead to a lot of information loss! Please check if you can collect better quality images!")
        compliant = self.__is_compliant(response=shapes)
        if compliant and not force_rescale:
            print("No rescaling needed!")
        else:
            # ...take each image and...
            for path in ["good", "bad"]:
                # ...create a new directory using the prefix
                os.mkdir("{}{}".format(prefix, path))
                files = os.listdir(path)
                # For each file...
                for file in files:
                    # ...load the image, resize it and save it back
                    # to the new folder
                    img_path = "{}/{}".format(path, file)
                    if not (".jpeg" in img_path or ".jpg" in img_path or ".png" in img_path):
                        print(
                            "The following image is not compliant: {}".format(img_path))
                        continue
                    image = io.imread(img_path)
                    rescaled = resize(image, shape_rec)
                    rescaled = img_as_ubyte(image=rescaled)
                    fname = "{}{}/{}".format(prefix, path, file)
                    io.imsave(fname=fname, arr=rescaled)
                # Log new image paths in output object
                output[path] = "{}{}".format(prefix, path)
        return output

    def upload_from_local(self, bucket, train_and_test=True, test_split=0.2, prefix="",
                          content_type="image/jpeg", processes_num=10):
        """This method will help you upload your local images to S3.
        Based on your folders "good" and "bad" - that are mandatory - it will
        upload the images to S3 accordingly.

        Args:
            bucket (str): The S3 bucket name
            train_and_test (bool): Have both present train and validation data
                Note: You can decide here wheter your Amazon Lookout for Vision
                project will use a train and validation set.
            test_split (float): The percentage to split between train and validation
            prefix (str): The prefix used for your "good" and "bad" folders.
                If you rescaled your new folders might be named "rescaled_good" and
                "rescaled_bad"
            content_type (str): The image type. Valid are "image/jpeg" and "image/png"

        Returns:
            json: Object with the S3 locations of the data.
        """ 
        # If you don't have folders "good" and "bad" this libary
        # will not work:
        folders = os.listdir()      
        if "{}good".format(prefix) not in folders and "{}bad".format(prefix) not in folders:
            print(
                "Error: Methods requires folders good/ and bad/ with images in this location!")
            return {}
        # For both folders "good" and "bad"...
        for p in ["{}good".format(prefix), "{}bad".format(prefix)]:
            files = os.listdir(path=p)
            # check if there are more than 5000 images in each "good" and "bad" folder, if so, exit
            check = self.__check_image_number(no_of_files=len(files))
            # shuffle and split the images into training and test set
            random.shuffle(files)
            training_set = files[:int(len(files)*(1-test_split))] 
            validation_set = files[-int(len(files)*test_split):]
            batches = [files]
            if train_and_test:
                batches = [training_set, validation_set]
                if "good" in p and len(training_set) < 10:
                    print("Warning: You don't have enough good images in your training dataset available for training a model!")
                    print("""
                        From the documentation:
                        A single dataset project needs at least 20 images labeled as normal and at least 10 images labeled as anomalous.
                        A project with separate training and test datasets, needs a training dataset with at least 10 images labeled as normal. The test dataset needs at least 10 images labeled as normal and at least 10 images labeled as anomalous.
                        We recommend that you add more than the minimum number of labeled images. Unlabeled images aren't used to train your model.
                    """)
                if "good" in p and len(validation_set) < 10:
                    print("Warning: You don't have enough good images in your validation dataset available for training a model!")
                    print("""
                        From the documentation:
                        A single dataset project needs at least 20 images labeled as normal and at least 10 images labeled as anomalous.
                        A project with separate training and test datasets, needs a training dataset with at least 10 images labeled as normal. The test dataset needs at least 10 images labeled as normal and at least 10 images labeled as anomalous.
                        We recommend that you add more than the minimum number of labeled images. Unlabeled images aren't used to train your model.
                    """)
                if "bad" in p and len(validation_set) < 10:
                    print("Warning: You don't have enough bad images in your validation dataset available for training a model!")
                    print("""
                        From the documentation:
                        A single dataset project needs at least 20 images labeled as normal and at least 10 images labeled as anomalous.
                        A project with separate training and test datasets, needs a training dataset with at least 10 images labeled as normal. The test dataset needs at least 10 images labeled as normal and at least 10 images labeled as anomalous.
                        We recommend that you add more than the minimum number of labeled images. Unlabeled images aren't used to train your model.
                    """)
            else:
                if "good" in p and len(files) < 20:
                    print("Warning: You don't have enough good images in your training dataset available for training a model!")
                    print("""
                        From the documentation:
                        A single dataset project needs at least 20 images labeled as normal and at least 10 images labeled as anomalous.
                        A project with separate training and test datasets, needs a training dataset with at least 10 images labeled as normal. The test dataset needs at least 10 images labeled as normal and at least 10 images labeled as anomalous.
                        We recommend that you add more than the minimum number of labeled images. Unlabeled images aren't used to train your model.
                    """)
                if "bad" in p and len(files) < 10:
                    print("Warning: You don't have enough bad images in your training dataset available for training a model!")
                    print("""
                        From the documentation:
                        A single dataset project needs at least 20 images labeled as normal and at least 10 images labeled as anomalous.
                        A project with separate training and test datasets, needs a training dataset with at least 10 images labeled as normal. The test dataset needs at least 10 images labeled as normal and at least 10 images labeled as anomalous.
                        We recommend that you add more than the minimum number of labeled images. Unlabeled images aren't used to train your model.
                    """)
            for batch_files in batches:
                subfolder = "normal"
                if "bad" in p:
                    subfolder = "anomaly"
                folder = "training"
                if batch_files == validation_set:
                    folder = "validation"
                # generate the input for batch uploading to s3 bucket
                batch_upload_input = []
                for file in batch_files:
                    filename = '{}/{}'.format(p, file)
                    if not (".jpeg" in file or ".jpg" in file or ".png" in file):
                        print("The following file was skipped: {}".format(file))
                        continue
                    input_tuple = (filename, bucket, "{}/{}/{}".format(folder, subfolder, file))
                    batch_upload_input.append(input_tuple)
                
                pool = ThreadPool(processes=processes_num)
                pool.starmap(self.__upload_s3, batch_upload_input)
        # Set metadata in response:
        response = {
            "train": {
                "normal": "s3://{}/training/normal/".format(bucket),
                "anomaly": "s3://{}/training/anomaly/".format(bucket)
            }
        }
        if train_and_test:
            response["test"] = {
                "normal": "s3://{}/validation/normal/".format(bucket),
                "anomaly": "s3://{}/validation/anomaly/".format(bucket)
            }
        return response

    def copy_from_s3(self, input_bucket, output_bucket, prefix_good="good", prefix_bad="bad",
                     train_and_test=True, test_split=0.2):
        """This method will help you copy images within S3.
        Based on your keys prefix_good and prefix_bad it will
        upload the images to S3 accordingly.

        Args:
            input_bucket (str): The S3 bucket name for input
            output_bucket (str): The S3 bucket name for output
            prefix_good (str): Path in S3 to good images
            prefix_bad (str): Path in S3 to bad images
            train_and_test (bool): Have both present train and validation data
                Note: You can decide here wheter your Amazon Lookout for Vision
                project will use a train and validation set.
            test_split (float): The percentage to split between train and validation

        Returns:
            json: Object with the S3 locations of the data.

        """
        s3 = boto3.client('s3')
        # Create a reusable Paginator
        paginator = s3.get_paginator('list_objects_v2')
        cnt = 0

        for p in [prefix_good, prefix_bad]:
            # Create a PageIterator from the Paginator
            page_iterator = paginator.paginate(Bucket=input_bucket, Prefix=p)
            # Paginate through S3
            for page in page_iterator:
                # For each key
                for key in page["Contents"]:
                    if key["Key"][-1] == "/":
                        continue
                    # only consider allowed images
                    if not (".jpeg" in key["Key"] or ".jpg" in key["Key"] or ".png" in key["Key"]):
                        print(
                            "The following file was skipped: {}".format(key["Key"]))
                        continue
                    # Then set the filename and depending on
                    # which folder you are in distinguish between
                    # normal ("good") and anomaly ("bad"):
                    subfolder = "normal"
                    if p == prefix_bad:
                        subfolder = "anomaly"
                    # Make a train-test-split:
                    choice = np.random.choice(a=[0, 1], size=1, replace=True, p=[
                                              1-test_split, test_split])[0]
                    folder = "training"
                    if choice == 1 and train_and_test:
                        folder = "validation"
                    # Set S3 object name and copy file:
                    fname = "{}/{}/{}".format(folder,
                                              subfolder, key["Key"].split("/")[-1])
                    
                    try:
                        
                        response = s3.copy({
                            "Bucket": input_bucket,
                            "Key": key["Key"]
                        }, output_bucket, fname)
                        cnt += 1
                    except Exception as e:
                        print("Warning: Copy did not work!")
                        print("With error message: {}".format(e))
        # Set metadata in response:
        response = {
            "status": "Success!",
            "objects_copied": cnt,
            "train": {
                "normal": "s3://{}/training/normal/".format(output_bucket),
                "anomaly": "s3://{}/training/anomaly/".format(output_bucket)
            }
        }
        if train_and_test:
            response["test"] = {
                "normal": "s3://{}/validation/normal/".format(output_bucket),
                "anomaly": "s3://{}/validation/anomaly/".format(output_bucket)
            }
        return response


# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
  
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
  
#       http://www.apache.org/licenses/LICENSE-2.0
  
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from datetime import datetime
import logging
import json
import boto3


class Manifest():
    """Helper class to assist with handling and creating manifest files.

    This class will assist you in generating your manifests. It also
    has a helper to push the manifest file to the correct location.

    Attributes:
    bucket          The name of the S3 bucket.
    datasets        The datasets used for the manifest (train and/or test)
    s3              The Amazon S3 boto3 client.

    """

    def __init__(self, bucket, datasets=["training"]):
        """Creates a Manifest helper object for Amazon Lookout for Vision.
        It can create a manifest file based on the data you stored on S3
        for your Amazon Lookout for Vision project. It can also push that
        artifact file to the correct place to get started with model training.

        Technical documentation on how Amazon Lookout for Vision works can be
        found at: https://aws.amazon.com/lookout-for-vision/

        Args:
            bucket (str): Name of the Amazon S3 bucket with your images.
            dataset (list): A list that needs at least "training" be present.
                Other option is "test".

        """
        super(Manifest, self).__init__()
        self.bucket = bucket
        self.datasets = datasets
        self.s3 = boto3.client("s3")

    def get_bucket(self):
        """Getter for bucket name.

        Args:
            None

        Returns:
            str: a S3 bucket name

        """
        return self.bucket

    def get_datasets(self):
        """Getter for datasets list.

        Args:
            None

        Returns:
            list: a list with the datasets to be used.

        """
        return self.datasets

    def generate_manifests(self):
        """Generates manifest file(s).
        It generates a manifest file for each option: training and test. If
        only training is specified for *datasets* then this function will only
        generate a manifest file for the training data stored in S3.

        Args:
            None

        Returns:
            json: a JSON object with one entry per image.

        """
        # Set variables:
        manifests = {}
        now = datetime.now()
        dttm = now.strftime("%Y-%m-%dT%H:%M:%S.%f")
        # For each dataset (namely training and/or test)...
        for ds in self.datasets:
            # ...add an empty JSON object to the manifests and...
            manifests[ds] = {}
            # ...for each folder of a dataset (needs to be 'normal' and 'anomaly')...
            for folder in ["normal", "anomaly"]:
                # ...list the objects with the prefix
                # e.g training/anomaly
                objects = self.s3.list_objects_v2(
                    Bucket=self.bucket, Prefix="{}/{}".format(ds, folder))["Contents"]
                # Manifest files assume labels to be:
                # 1 = good image
                # 0 = bad image
                label = 1
                if folder == "anomaly":
                    label = 0
                # Each manifest will be represented as a string:
                manifest = ""
                # For each file from the S3 objects...
                for file in objects:
                    # ...extract the filename and...
                    f = file["Key"].split("/")[-1]
                    # ...set the manifest object as per:
                    manifest_obj = {
                        "source-ref": "s3://{}/{}/{}/{}".format(self.bucket, ds, folder, f),
                        "auto-label": label,
                        "auto-label-metadata": {
                            "confidence": 1,
                            "job-name": "labeling-job/auto-label",
                            "class-name": folder,
                            "human-annotated": "yes",
                            "creation-date": dttm,
                            "type": "groundtruth/image-classification"
                        }
                    }
                    # Add a dump of each object to the manifest
                    manifest += (json.dumps(manifest_obj)+"\n")
                # Once a manifest is created add it to all manifests:
                manifests[ds][folder] = manifest
        return manifests

    def push_manifests(self):
        """Generates and push manifest file(s).
        It will use the generate_manifests(...) method to generate the
        manifest files and push them to S3.

        Args:
            None

        Returns:
            json: a JSON object either empty (failure) or the S3 locations
                of the manifest files

        """
        # Generate manifest files
        manifests = self.generate_manifests()
        success = {}
        # Try...
        try:
            # for each dataset (namely training and/or test)...
            for ds in self.datasets:
                # ...each manifest will be represented as a string
                manifest = ""
                for key in manifests[ds]:
                    # Add a dump of each object to the manifest
                    manifest += manifests[ds][key]
                # Upload the manifest to S3
                upload = self.s3.put_object(
                    Bucket=self.bucket, Key="{}.manifest".format(ds), Body=manifest)
                # Store location
                success[ds] = {
                    "bucket": self.bucket,
                    "key": "{}.manifest".format(ds),
                    "location": "s3://{}/{}.manifest".format(self.bucket, ds)}
        except Exception as e:
            # Log error!
            logging.error(e)
        return success


from setuptools import find_packages, setup

with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='lookoutvision',
    packages=find_packages(include=['lookoutvision']),
    version='0.1.0',
    description='Python SDK for Amazon Lookout for Vision',
    author='Michael Wallner',
    author_email="wallnm@amazon.com",
    license='Apache-2.0',
    install_requires=[
        'numpy',
        'pandas',
        'boto3',
        'scikit-image',
        'seaborn',
        'matplotlib'
    ],
    url='',
    setup_requires=['pytest-runner'],
    tests_require=['pytest==4.4.1'],
    test_suite='tests',
    zip_safe=True)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
  
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
  
#       http://www.apache.org/licenses/LICENSE-2.0
  
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os
import boto3
from collections import defaultdict
import inspect
import json
import logging
import time


class LookoutForVision():
    """LookoutForVision class to build, train and deploy.

    This class helps to build, train and deploy a Amazon Lookout for Vision
    project. It implements the three most common methods for model deployment:
    # - .fit()
    # - .deploy()
    # - .predict()

    Attributes:
    project_name    The name of the Amazon Lookout for Vision project.
    lv              The Amazon Lookout for Vision boto3 client.
    model_version   The (initial) model version.

    """

    def __init__(self, project_name, model_version="1"):
        """Build, train and deploy Amazon Lookout for Vision models.

        Technical documentation on how Amazon Lookout for Vision works can be
        found at: https://aws.amazon.com/lookout-for-vision/

        Args:
            project_name (str): Name of the Amazon Lookout for Vision to interact with.
            model_version (str): The (initial) model version.

        """
        #super(LookoutForVision, self).__init__()
        self.project_name = project_name
        self.lv = boto3.client("lookoutvision")
        self.s3 = boto3.client("s3")
        self.model_version = model_version
        self.describe_project()

    @classmethod
    def _get_param_names(self):
        """Internal get parameter names helper.
        It will retrieve all the parameters used within your class.

        Args:
            None

        Returns:
            list: all class parameters

        """
        init = getattr(self.__init__, 'deprecated_original', self.__init__)
        init_signature = inspect.signature(init)
        parameters = [p for p in init_signature.parameters.values()
                      if p.name != 'self' and p.kind != p.VAR_KEYWORD]
        return sorted([p.name for p in parameters])

    def describe_project(self):
        """Describe a project.

        Args:
            None

        Returns:
            json: The project details

        """

        project = {}
        # First try to describe the project given by the name:
        try:
            project = self.lv.describe_project(
                ProjectName=self.project_name)["ProjectDescription"]["ProjectArn"]
            print('Project already exists with arn: ' + project)
        except Exception as e:
            if 'ResourceNotFoundException' in str(e):
                print(f"Project {self.project_name} does not exist yet...use the create_project() method to set up your first project")
            else:
                raise Exception
        return project

    def create_project(self):
        """Create a project.

        Args:
            None

        Returns:
            json: The project details

        """
        project = {}
        # First try to create a new project:
        try:
            project = self.lv.create_project(
                ProjectName=self.project_name
            )["ProjectMetadata"]["ProjectArn"]
            print(f"Creating the project: {self.project_name}")
        except Exception as e:
            if 'ConflictException' in str(e):
                project = self.lv.describe_project(
                    ProjectName=self.project_name
                )["ProjectDescription"]["ProjectArn"]
            else:
                raise Exception
        return project

    def get_params(self, deep=True):
        """Get class parameters.

        Args:
            deep (bool): Make a deep copy of parameters for output.

        Returns:
            json: an object with the internal parameters and their values.

        """
        output = dict()
        for key in self._get_param_names():
            value = getattr(self, key)
            if deep and hasattr(value, 'get_params'):
                deep_items = value.get_params().items()
                output.update((key + '__' + i, val) for i, val in deep_items)
            output[key] = value
        return output

    def set_params(self, **params):
        """Set class parameters.

        Args:
            **params (dict): New parameters in key=value format.

        Returns:
            self: the class itself

        """
        if not params:
            return self
        valid = self.get_params(deep=True)

        nested = defaultdict(dict)
        for key, value in params.items():
            key, delim, sub = key.partition('__')
            if key not in valid:
                raise ValueError('Invalid parameter %s for class %s. '
                                 'Check the list of available parameters '
                                 'with `cls.get_params().keys()`.' %
                                 (key, self))
            if delim:
                nested[key][sub] = value
            else:
                setattr(self, key, value)
                valid[key] = value

        for key, sub_params in nested.items():
            valid[key].set_params(**sub_params)

        return self

    def update_datasets(self, dataset_type, wait=True):
        """Create a dataset.

        Args:
            dataset_type (dict): A setting from where to get input data from
                Format of this input is of type:
                "train": {
                    "bucket": "my_s3_bucket",
                    "key": "training.manifest",
                    "version": "1"
                },
                "test": {
                    "bucket": "my_s3_bucket",
                    "key": "validation.manifest",
                    "version": "1"
                }
            wait (bool): Either to wait in the console uppon succes or escape function

        Returns:
            json: an object with metadata on success

        """

        datasets = {}
        # For each dataset used...
        for key in dataset_type:
            # ...create a dataset
            d_type = "train" if (
                key == "training" or key == "train") else "test"
            try:
                dataset = self.lv.delete_dataset(
                    ProjectName=self.project_name,
                    DatasetType=d_type
                )
            except Exception as e:
                print("Error in dataset deletion with exception: {}".format(e))
                print("Please check CloudWatch logs for troubleshooting!")
        return self.create_datasets(dataset_type=dataset_type, wait=wait)

    def create_datasets(self, dataset_type, wait=True):
        """Create a dataset.

        Args:
            dataset_type (dict): A setting from where to get input data from
                Format of this input is of type:
                "train": {
                    "bucket": "my_s3_bucket",
                    "key": "training.manifest",
                    "version": "1"
                },
                "test": {
                    "bucket": "my_s3_bucket",
                    "key": "validation.manifest",
                    "version": "1"
                }
            wait (bool): Either to wait in the console uppon succes or escape function

        Returns:
            json: an object with metadata on success

        """
        datasets = {}
        # For each dataset used...
        for key in dataset_type:
            # ...create a dataset
            d_type = "train" if (
                key == "training" or key == "train") else "test"
            try:
                dataset = self.lv.create_dataset(
                    ProjectName=self.project_name,
                    DatasetType=d_type,
                    DatasetSource={
                        'GroundTruthManifest': {
                            'S3Object': {
                                'Bucket': dataset_type[key]["bucket"],
                                'Key': dataset_type[key]["key"],
                                'VersionId': self.model_version
                            }
                        }
                    }
                )["DatasetMetadata"]
                # Log output
                datasets[key] = dataset
            except Exception as e:
                if 'ConflictException' in str(e):
                    print("Dataset already existed in the project")
                    print(
                        "If the dataset already exists try updating it with: update_datasets")
                else:    
                    print("Error in create_datasets with exception: {}".format(e))
                    raise Exception
                return datasets
        # Notify user when creation is done:
        print("Creating dataset(s):", end=" ")
        if wait:
            status = "CREATING"
            while status != "CREATE_COMPLETE":
                cnt = 0
                for key in dataset_type:
                    d_type = "train" if (
                        key == "training" or key == "train") else "test"
                    stat = self.lv.describe_dataset(
                        ProjectName=self.project_name,
                        DatasetType=d_type
                    )["DatasetDescription"]["Status"]
                    if stat == "CREATE_COMPLETE":
                        cnt += 1
                        datasets[key]["Status"] = "CREATE_COMPLETE"
                if cnt == len(dataset_type):
                    status = "CREATE_COMPLETE"
                print("-", end="")
                time.sleep(5)
            print("!")
        for key in datasets:
            datasets[key]["StatusMessage"] = "Dataset created."
        return datasets

    def fit(self, output_bucket, model_prefix=None, train_and_test=True, wait=True):
        """Train the model.
        Create a model from the datasets
        At first check whether the minimum no of images are available to train the model.
        There should be min 20 normal and 10 anomalous images in training/train dataset.

        Args:
            output_bucket (str): The output S3 bucket to be used for model logging.
            model_prefix (str): Optional to add a model prefix name for logging.
            train_and_test (bool): Whether to us train or train and test set
            wait (bool): Either to wait in the console uppon succes or escape function

        Returns:
            json: an object with metadata on success

        """

        test_dataset = {"Status": "No test dataset used!"}
        if train_and_test:
            test_dataset = self.lv.describe_dataset(
                ProjectName=self.project_name,
                DatasetType="test"
            )["DatasetDescription"]
        train_dataset = self.lv.describe_dataset(
            ProjectName=self.project_name,
            DatasetType="train"
        )["DatasetDescription"]
        normal_no_images_train = train_dataset["ImageStats"]["Normal"]
        anomaly_no_images_train = train_dataset["ImageStats"]["Anomaly"]

        if ((normal_no_images_train >= 20 and anomaly_no_images_train >= 10)):
            model = self.lv.create_model(
                ProjectName=self.project_name,
                OutputConfig={
                    'S3Location': {
                        'Bucket': output_bucket,
                        'Prefix': model_prefix if model_prefix is not None else ""
                    }
                })["ModelMetadata"]
            if wait:
                print("Model training started:", end=" ")
                version = model["ModelVersion"]
                status = model["Status"]
                while status == "TRAINING":
                    update = self.lv.describe_model(
                        ProjectName=self.project_name,
                        ModelVersion=version
                    )["ModelDescription"]
                    status = update["Status"]
                    print("-", end="")
                    time.sleep(60)
                print("!")
            else:
                print("""Model is being created. Training will take a while.\n
                         Please check your Management Console on progress.\n
                         You can continue with deployment once the model is trained.\n
                      """)
            # Return success
            return {
                "status": "Success!",
                "project": self.project_name,
                "train_datasets": train_dataset,
                "test_datasets": test_dataset,
                "model": model
            }
        else:
            print("""Number of images is not sufficient, at least 20 normal and 10 anomaly\n
                     imgages are required for training images
                  """)
            return {
                "status": "Failure!",
                "project": self.project_name,
                "train_datasets": train_dataset,
                "test_datasets": test_dataset,
                "model": None
            }

    def deploy(self, min_inf_units=1, model_version=None, wait=True):
        """Deploy your model.

        Args:
            min_inf_units (int): Minimal number of inference units.
            model_version (str): The model version to deploy.
            wait (bool): Either to wait in the console uppon succes or escape function

        Returns:
            json: an object with metadata on success

        """
        # Check the current status of the model
        current_status = self.lv.describe_model(
            ProjectName= self.project_name,
            ModelVersion=self.model_version if model_version is None else model_version
        )['ModelDescription']['Status']
        # If model status is TRAINED , then only start the model. 
        # otherwise print the message with actual status that it can not be started
        if (current_status != 'TRAINED'):
            print('current model with version {} is in the status {}, hence it can not be started/hosted. The model needs to be in TRAINED status to be started/hosted'.format(self.model_version if model_version is None else model_version,current_status))
        else:
            # Start the model either using the internal model version
            # or use the one supplied to this method:    
            status = self.lv.start_model(
                ProjectName=self.project_name,
                ModelVersion=self.model_version if model_version is None else model_version,
                MinInferenceUnits=min_inf_units
            )["Status"]
            # Wait until model is trained:
            print("Model will be hosted now")
            if wait:
                while status != "HOSTED":
                    status = self.lv.describe_model(
                        ProjectName=self.project_name,
                        ModelVersion=self.model_version if model_version is None else model_version
                    )["ModelDescription"]["Status"]
                    print("-", end="")
                    time.sleep(60)
                print("!")
                print("Your model is now hosted!")
            # Return success:
            return {
                "status": status,
                "model": self.lv.describe_model(
                    ProjectName=self.project_name,
                    ModelVersion=self.model_version if model_version is None else model_version
                )["ModelDescription"]
            }
        
    def predict(self, model_version=None, local_file="",
                bucket="", key="", content_type="image/jpeg"):
        """Predict using your Amazon Lookout for Vision model.
        You can either predict from S3 object or local images.

        Args:
            model_version (str): The model version to deploy.
            local_file (str): Path to local image.
            bucket (str): S3 bucket name.
            key (str): Object in S3 bucket.
            content_type (str): Either "image/jpeg" or "image/png".

        Returns:
            json: an object with results of prediction

        """
        # If no paths are set return warning:
        if local_file == "" and bucket == "" and key == "":
            print("Warning: either local_file OR bucket & key need to be present!")
            return {'Source': {'Type': 'warning'}, 'IsAnomalous': None, 'Confidence': -1.0}
        # If paths for local file AND S3 bucket are set return another warning:
        if local_file != "" and bucket != "" and key != "":
            print("Warning: either local_file OR bucket & key need to be present!")
            return {'Source': {'Type': 'warning'}, 'IsAnomalous': None, 'Confidence': -1.0}
        # If method is used properly then...
        obj = None
        if local_file != "":
            # ...set obj to bytearray using local image...
            try:
                with open(local_file, "rb") as image:
                    f = image.read()
                    obj = bytearray(f)
            except IsADirectoryError as e:
                print("Warning: you specified a directory, instead of a single file path")
                print("Maybe you would like to try \'_batch_predict_local\' or \'batch_predict_s3\' method!")
                return
        elif bucket != "" and key != "":
            # ...or a byte object by pulling from S3
            obj = boto3.client("s3").get_object(
                Bucket=bucket,
                Key=key)["Body"].read()
        else:
            # If file not found:
            print("Warning: No file found!")
            return {'Source': {'Type': 'warning'}, 'IsAnomalous': None, 'Confidence': -1.0}
        # Predict using your model:
        result = self.lv.detect_anomalies(
            ProjectName=self.project_name,
            ModelVersion=self.model_version if model_version is None else model_version,
            Body=obj,
            ContentType=content_type
        )["DetectAnomalyResult"]
        return result

    def _batch_predict_local(self, local_path, model_version=None, content_type="image/jpeg"):
        """Predict for all the images using your Amazon Lookout for Vision model 
        from S3 objects.

        Args:
            local_path (str): Path to local images.
            model_version (str): The model version to deploy.
            content_type (str): Either "image/jpeg" or "image/png".

        Returns:
            json: s3 objects with location which stores the results of prediction

        """
        predictions = []
        files = os.listdir(local_path)
        for file in files:
            filename = '{}/{}'.format(local_path, file)
            # ...set obj to bytearray using local image...
            with open(filename, "rb") as image:
                f = image.read()
                obj = bytearray(f)
                try:
                    # Predict using your model:
                    result = self.lv.detect_anomalies(
                        ProjectName=self.project_name,
                        ModelVersion=self.model_version if model_version is None else model_version,
                        Body=obj,
                        ContentType=content_type
                    )["DetectAnomalyResult"]
                    predictions.append(result)
                except Exception as e:
                    print("Warning: prediction failed for file: {}".format(filename))
                    print("with error message: {}".format(e))
        return {'status': 'Success!', 'predicted_result': predictions}

    def batch_predict(self, model_version=None, local_path=None, input_bucket="", input_prefix=None,
                      output_bucket="", output_prefix=None, content_type="image/jpeg"):
        """Predict for all the images using your Amazon Lookout for Vision model 
        from S3 objects.

        Args:
            model_version (str): The model version to deploy.
            local_path (str): Path to local images.
            input_bucket (str): S3 bucket name for input images.
            input_prefix(str): S3 folder names ( if any) for input bucket location
            output_bucket (str): S3 bucket name to store predicted results.
            output_prefix(str): S3 folder names ( if any) for output bucket location to store predicted results
            content_type (str): Either "image/jpeg" or "image/png".

        Returns:
            json: s3 objects with location which stores the results of prediction

        """
        if local_path is not None:
            return self._batch_predict_local(local_path=local_path,
                                             model_version=model_version, content_type=content_type)
            # If no input bucket is set return warning:
        if input_bucket == "":
            print("Warning: S3 bucket need to be present for input images!")
            return {'status': 'Error!', 'predicted_result': None}
        # If no output bucket is set return warning:
        if output_bucket == "":
            print("Warning: S3 bucket need to be present to load prediction result!")
            return {'status': 'Error!', 'predicted_result': None}
        obj = None
        if input_bucket != "":
            
            success = {}
            kwargs = {'Bucket': input_bucket}
            if (input_prefix != None):
                if isinstance(input_prefix, str):
                    kwargs['Prefix'] = input_prefix
                else:
                    kwargs['Prefix'] = str(input_prefix)
            paginator = self.s3.get_paginator('list_objects_v2')
            pages = paginator.paginate(**kwargs)

            # Try...
            try:
            
                for page in pages:
                    for obj in page['Contents']:           
                
                        key = obj['Key']
                        if key[-1] == "/":
                            continue
                    
                        body = self.s3.get_object(Bucket=input_bucket, Key=key)[
                            "Body"].read()
                        file_name = key.split('/')[-1]
                    # Predict using your model:
                        result = self.lv.detect_anomalies(
                            ProjectName=self.project_name,
                            ModelVersion=self.model_version if model_version is None else model_version,
                            Body=body,
                            ContentType=content_type)["DetectAnomalyResult"]
                    # Upload the manifest to S3
                        upload = self.s3.put_object(Bucket=output_bucket, Key=output_prefix+"{}.json".format(
                            file_name), Body=json.dumps(result), ServerSideEncryption='AES256')
                        success = {"output_bucket": output_bucket,
                                   "prdected_file_key": output_prefix+"{}.json".format(file_name)
                               
                                   }
                        print('Predicted output is uploaded to s3 :' +
                              json.dumps(success))
                    
            except Exception as e:
                logging.error(e)
                print('Key object corresponding to error :' + key)
                       
                        
            return {
                'status': 'Success!',
                'predicted_result': "s3://{}/{}".format(output_bucket, output_prefix)
            }

        else:
            # If file not found:
            print("Warning: No file found!")
            return {'status': 'Error!', 'predicted_result': None}

    def stop_model(self, model_version=None, wait=True):
        """Stop deployed model.

        Args:
            model_version (str): The model version to deploy.

        Returns:
            json: an object with results of prediction

        """
        response = {}
        try:
            # Stop the model
            ModelVersion = self.model_version if model_version is None else model_version
            print('Stopping model version ' + ModelVersion +
                  ' for project ' + self.project_name)
            response = self.lv.stop_model(ProjectName=self.project_name,
                                          ModelVersion=ModelVersion)
            status = response["Status"]
            print("Model will be stopped now")
            if wait:
                while status != "TRAINED":
                    status = self.lv.describe_model(
                        ProjectName=self.project_name,
                        ModelVersion=ModelVersion
                    )["ModelDescription"]["Status"]
                    print("-", end="")
                    time.sleep(5)
                print("!")
                print("Your model is now stopped!")
            print('Status: ' + response['Status'])
        except Exception as e:
            response["Error"] = e
            print("Something went wrong: ", e)
        return response
